from cpbox.tool import redistool
from cpbox.tool import local_cache

FROM_LOCAL_CACHE = 1
FROM_REDIS = 2
FROM_SLOW_OP = 3

class CacheUnit(object):

    def __init__(self, key, get_fn, expire=3600):
        self.key = key
        self.get_fn = get_fn
        self.expire = expire

    def fetch(self):
        return fetch(self.key, self.get_fn, None, None, self.expire)

    def invalid(self):
        delete(self.key)

def set_redis_url(url):
    redistool.init_proxy(url)

def delete(key):
    local_cache.delete(key)
    redistool.delete(key)

def fetch(key, get_fn, on_to_local_fn=False, return_filter_fn=False, expire=3600):
    get_key_fn = lambda key:key
    get_list_fn = lambda _ids:{key: get_fn(key)}
    return fetch_list(key, get_key_fn, get_list_fn, on_to_local_fn, return_filter_fn, expire)

def _get_list_from_cache_and_local_cache(keys, use_local_cache=True, on_to_local_fn=None, return_from_tag=False):
    list = {}
    un_hit_keys = []
    if use_local_cache:
        for key in keys:
            if local_cache.has(key):
                list[key] = (local_cache.get(key), FROM_LOCAL_CACHE) if return_from_tag else local_cache.get(key)
            else:
                un_hit_keys.append(key)
    else:
        un_hit_keys = keys

    if un_hit_keys:
        list_from_cache = redistool.get_list(un_hit_keys)
        for key, item in list_from_cache.items():
            if callable(on_to_local_fn):
                item = on_to_local_fn(key, item)
            list[key] = (item, FROM_REDIS) if return_from_tag else item
            local_cache.set(key, item)
    return list

def _fetch_list_from_cache(id_to_key_map, on_to_local_fn=None, use_local_cache=True, return_from_tag=False):
    keys = id_to_key_map.values()
    if callable(on_to_local_fn):
        key_to_id_map = {v: k for k, v in id_to_key_map.items()}
        def on_to_local_fn_proxy(key, item):
            id = key_to_id_map[key]
            return on_to_local_fn(id, item)
        return _get_list_from_cache_and_local_cache(keys, use_local_cache, on_to_local_fn_proxy, return_from_tag)
    else:
        return _get_list_from_cache_and_local_cache(keys, use_local_cache, None, return_from_tag)

def set_to_cache_for_fetch(id, raw_item, key=None, get_key_fn=None, on_to_local_fn=None, expire=600):
    if key is None and get_key_fn is None:
        raise Error('key or get_key_fn must be defined')

    if not key:
        key = get_key_fn(id)

    item = raw_item
    if on_to_local_fn:
        item = on_to_local_fn(id, raw_item)
    if expire > 0:
        redistool.set(key, raw_item, expire)
        local_cache.set(key, item, expire)
    return item

def fetch_list(ids, get_key_fn, get_list_fn, on_to_local_fn=None, return_filter_fn=None, expire=600, return_from_tag=False):
    '''
    1. if expire >= 0, will try to use local cache and redis;
       the raw result generated by get_list_fn will be cached to both cache.

    2. if expire == 0, will not fetch from local cache
    '''
    if not ids:
        return []

    is_list = isinstance(ids, list)
    if not is_list:
        ids = [ids]

    id_to_key_map = {}
    for id in ids:
        id_to_key_map[id] = get_key_fn(id)

    cache_list = {}
    if expire >= 0:
        cache_list = _fetch_list_from_cache(id_to_key_map, on_to_local_fn, expire > 0, return_from_tag)

    un_hit_ids = []
    result_map = {}
    for id, key in id_to_key_map.items():
        if key in cache_list:
            result_map[id] = cache_list[key]
        else:
            un_hit_ids.append(id)

    if un_hit_ids:
        raw_result = get_list_fn(un_hit_ids)
        for id, item in raw_result.items():
            key = id_to_key_map[id]
            item = set_to_cache_for_fetch(id, item, key=key, on_to_local_fn=on_to_local_fn, expire=expire)
            result_map[id] = (item, FROM_SLOW_OP) if return_from_tag else item

    filtered_result = {}
    if callable(return_filter_fn):
        for id, item in result_map.items():
            item = return_filter_fn(id, item)
            if not item:
                filtered_result[id] = item
    else:
        filtered_result = result_map

    if is_list:
        return filtered_result
    else:
        values = filtered_result.values()
        if values:
            return next(iter(values))
        else:
            return False



from os import path
import unittest

from cpbox.tool import file
from cpbox.tool import utils
from cpbox.tool.testutils import RandomDir

class TestUtils(unittest.TestCase, RandomDir):

    def setUp(self):
        RandomDir.setUp(self)
        self.template_dir = path.join(self.tests_root_dir, 'test-data')

    def tearDown(self):
        RandomDir.tearDown(self)

    def test_properties(self):
        dst_filename = path.join(self.temp_dir, 'out.properties')
        src_filename = path.join(self.template_dir, 'test-utils-properties.txt')
        ret = utils.load_properties(src_filename, to_dict=False)
        self.assertEqual(ret['VERSION_CODE'], '989')
        utils.dump_properties(ret, dst_filename)
        self.assertEqual(file.file_get_contents(dst_filename).strip(), file.file_get_contents(src_filename).strip())

    def test_yaml(self):
        src_filename = path.join(self.template_dir, 'test.yaml')
        dst_filename = path.join(self.temp_dir, 'test.yaml')
        ret = utils.load_yaml(src_filename)
        self.assertEqual(ret['BODY_0']['Mesh_File'], 'taylor_impact.dat')
        utils.dump_yaml(dst_filename, ret)
        ret1 = utils.load_yaml(src_filename)
        self.assertEqual(ret, ret1)

if __name__ =='__main__':
    unittest.main()



import getpass
import os

from cpbox.app.devops import DevOpsApp
from cpbox.app.devops import DevOpsAppConfigProvider
from cpbox.tool import file
from cpbox.tool import functocli
from cpbox.tool import utils


class SimpleMultiDeveloperDevOpsAppConfigProvider(DevOpsAppConfigProvider):

    def __init__(self, app, app_base_name, kwargs):
        self.app = app
        self.app_base_name = app_base_name
        self.kwargs = kwargs
        self.developer_name = self._determine_developer_name()
        app_name = self.determine_app_name()
        DevOpsAppConfigProvider.__init__(self, app_name)

    def _determine_developer_name(self):
        app_config = DevOpsAppConfigProvider.get_box_local_config(self)
        developer_name = app_config.get('developer_name', None)
        if not developer_name:
            developer_name = getpass.getuser()
        developer_name_from_cmd = self.kwargs.get('developer_name', None)
        developer_name = developer_name_from_cmd if developer_name_from_cmd else developer_name
        return developer_name

    def determine_app_name(self):
        if self.app.isolate_by_developer():
            env = self.get_env()
            app_name = '%s-%s-%s' % (env, self.developer_name, self.app_base_name)
            return app_name
        else:
            return self.app_base_name

    def get_app_root_dir(self):
        app_root_dir = self.get_roles_dir() / self.app_base_name
        return app_root_dir

    def get_env(self):
        env_from_puppy_env = DevOpsAppConfigProvider.get_env(self)
        env_from_cmd_args = self.kwargs.get('env')
        return env_from_cmd_args if env_from_cmd_args else env_from_puppy_env


class MultiDeveloperDevOpsAppConfigProvider(SimpleMultiDeveloperDevOpsAppConfigProvider):

    def __init__(self, app, app_base_name, kwargs):
        self.sandbox_code_dir = None
        SimpleMultiDeveloperDevOpsAppConfigProvider.__init__(self, app, app_base_name, kwargs)
        self._process_sandbox()

    def get_root_dir(self):
        if self.sandbox_code_dir is not None:
            return self.sandbox_code_dir
        else:
            return SimpleMultiDeveloperDevOpsAppConfigProvider.get_root_dir(self)

    def get_source_code_root_dir(self):
        return SimpleMultiDeveloperDevOpsAppConfigProvider.get_root_dir(self)

    def _process_sandbox(self):
        if self.sandbox_version != 0:
            self.sandbox_code_dir = '%s/sandbox-code-%03d' % (self.app_persistent_storage_dir, self.sandbox_version)

    def determine_app_name(self):
        env = self.get_env()
        if env == 'test' or env == 'dev':
            self.sandbox_version = self.kwargs.get('sandbox_version', 0)
        else:
            self.sandbox_version = 0

        self.shadow_id = int(os.getenv('shadow_id', 0))

        app_name = '%s-%s-%s' % (env, self.developer_name, self.app_base_name)
        if self.sandbox_version != 0:
            app_name = '%s-%s-%s-sandbox-%03d' % (env, self.developer_name, self.app_base_name, self.sandbox_version)

        if self.shadow_id != 0:
            app_name = '%s-%s-%s-shadow-%03d' % (env, self.developer_name, self.app_base_name, self.shadow_id)
        return app_name


class SimpleMultiDeveloperDevOpsApp(DevOpsApp):

    def __init__(self, app_base_name, **kwargs):
        self.app_base_name = app_base_name
        self.kwargs = kwargs

        self.app_config_provider = SimpleMultiDeveloperDevOpsAppConfigProvider(self, self.app_base_name, self.kwargs)
        self.developer_name = self.app_config_provider.developer_name

        DevOpsApp.__init__(self, self.app_config_provider.app_name, **kwargs)

    def create_app_config_provider(self):
        return self.app_config_provider

    def isolate_by_developer(self):
        return getpass.getuser() != 'root'

    @staticmethod
    def run_app(app, log_level='info', default_method=None, common_args_option=None):
        common_args_option_basic = {
            'args': ['env', 'developer_name'],
            'default_values': {
                'env': None,
                'developer_name': None,
            }
        }
        functocli.run_app(app, log_level, default_method, common_args_option, common_args_option_basic)


class MultiDeveloperDevOpsApp(DevOpsApp):

    def __init__(self, app_base_name, **kwargs):
        self.app_base_name = app_base_name
        self.kwargs = kwargs

        self.app_config_provider = MultiDeveloperDevOpsAppConfigProvider(self, self.app_base_name, self.kwargs)
        self.developer_name = self.app_config_provider.developer_name

        DevOpsApp.__init__(self, self.app_config_provider.app_name, **kwargs)

        self.shadow_id = self.app_config_provider.shadow_id
        self.sandbox_version = self.app_config_provider.sandbox_version

        if self.is_dev():
            self.do_init_for_dev(False)
        else:
            self._prepare_config_for_none_dev()

    def isolate_by_developer(self):
        return getpass.getuser() != 'root'

    def create_app_config_provider(self):
        return self.app_config_provider

    def is_dev(self):
        env = self.app_config_provider.get_env()
        return env == 'dev' or env == 'test'

    def _prepare_config_for_none_dev(self):
        data = utils.load_yaml(self.app_config_dir / 'env-config.yml')
        self.env_config = data[self.env]
        self.app_init_config_by_developer_name()
        self.app_do_init_for_deploy()

    @functocli.keep_method
    def do_init_for_dev(self, force=True):
        if self.sandbox_version:
            self.check_out_for_sandbox()
        self.app_config_provider.ensure_dir_and_write_permission()

        self.dev_env_config_file = self.app_persistent_storage_dir / 'dev-env-config.yml'
        if not os.path.isfile(self.dev_env_config_file) or self.shadow_id or force:
            config = self.app_build_env_config()
            utils.dump_yaml(self.dev_env_config_file, config)
            self.logger.info('generate dev env config: %s', self.dev_env_config_file)
            self.env_config = config
            self.app_init_config_by_developer_name()
            self.app_do_init_for_dev()
        else:
            self.env_config = utils.load_yaml(self.dev_env_config_file)
            self.app_init_config_by_developer_name()

    def check_out_for_sandbox(self):
        provider = self.app_config_provider
        file.ensure_dir(provider.sandbox_code_dir)
        cmd = "rsync -xa --no-links --delete --exclude='.git' --filter='dir-merge,- .gitignore' %s/ %s" % (provider.get_source_code_root_dir(), provider.get_root_dir())
        self.shell_run(cmd)

    def app_init_config_by_developer_name(self):
        raise NotImplementedError("Please implement this method")

    def app_do_init_for_deploy(self):
        raise NotImplementedError("Please implement this method")

    def app_do_init_for_dev(self):
        raise NotImplementedError("Please implement this method")

    def app_build_env_config(self):
        raise NotImplementedError("Please implement this method")

    @staticmethod
    def run_app(app, log_level='info', default_method=None, common_args_option=None):
        common_args_option_basic = {
            'args': ['env', 'sandbox_version', 'developer_name'],
            'default_values': {
                'env': None,
                'sandbox_version': 0,
                'developer_name': None,
            }
        }
        functocli.run_app(app, log_level, default_method, common_args_option, common_args_option_basic)



from cpbox.tool import datatypes


class LegacySpec(object):

    def __init__(self, def_list, config_name, kv=None):
        self.def_list = def_list
        self.config_name = config_name
        self.config_names = set([self.config_name])

        self.config_data = {}
        if def_list:
            self.config_data_raw = {}
            if config_name in self.def_list:
                self.config_data_raw = self.def_list[config_name]

            self.config_data_parent = {}
            config_data = dict(self.config_data_raw)
            if kv:
                config_data.update({item[0]: item[1] for item in (_kv.split('=') for _kv in kv.split(','))})
            self.config_data = config_data

            if 'parent_config' in config_data:
                parent_config = config_data.get('parent_config', ())
                del config_data['parent_config']
                self._process_parent_config(parent_config)

    def _process_parent_config(self, parent_config):
        config_data = {}
        for _name in parent_config:
            # overwrite config
            parent_ecs_def = self.__class__(self.def_list, _name)
            self.config_names.update(parent_ecs_def.config_names)
            config_data = datatypes.merge(config_data, parent_ecs_def.config_data, 'merge')
        self.config_data_parent = dict(config_data)

        self.config_data = datatypes.merge(config_data, self.config_data, 'merge')

    def __str__(self):
        data = self.config_data
        return str(data)

    def __repr__(self):
        return self.__str__()



import unittest
from cpbox.tool import concurrent
from cpbox.tool import logger
from threading import Thread
import time
import logging

log = logging.getLogger()


class DelayBatchOpTest(object):

    def __init__(self):
        self.added_num = 0
        self.processed_num = 0
        self.callback_items = []

    def run(self):
        num_max = 25
        put_item_interval = 0.001
        timeout = 0.01
        batch_size = 10

        def callback(items):
            self.callback_items.extend(items)
            self.processed_num += len(items)

        def wait():
            while len(self.callback_items) != num_max:
                time.sleep(put_item_interval * 3)

        def add_element():
            while self.added_num < num_max:
                time.sleep(put_item_interval)
                self.added_num = self.added_num + 1
                delay_op.add_delay_item(self.added_num)

        delay_op = concurrent.DelayBatchOp(callback, timeout, batch_size=batch_size)
        thread_add_element = Thread(target=add_element)
        thread_add_element.daemon = True
        thread_add_element.start()

        thread_wait = Thread(target=wait)
        thread_wait.daemon = True
        thread_wait.start()
        thread_wait.join()


class TRepeatedTimerTest:

    def __init__(self):
        self.processed_num = 0
        self.max_count = 25

    def run(self):
        interval = 0.001

        def callback():
            self.processed_num += 1

        timer = concurrent.TRepeatedTimer(callback, interval, max_count=self.max_count)
        timer.start()
        timer.join()


class TestConcurrent(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_deplay_batch_op(self):
        test = DelayBatchOpTest()
        test.run()
        self.assertEqual(test.processed_num, test.added_num)

    def test_thread_base_timer(self):
        test = TRepeatedTimerTest()
        test.run()
        self.assertEqual(test.processed_num, test.max_count)


if __name__ == '__main__':
    logger.make_logger_for_test()
    unittest.main()



import socket
import psutil

def get_ip_address(ifname):
    addrs = psutil.net_if_addrs()
    if ifname in addrs:
        for addr in addrs[ifname]:
            if addr.family == socket.AF_INET:
                return addr.address
    raise ValueError(f"No IPv4 address found for interface {ifname}")

def get_hostname_ip():
    ip = socket.gethostbyname(socket.gethostname())
    return ip

def get_local_ip_address():
    try:
        return get_ip_address_udp()
    finally:
        return '127.0.0.1'

def get_ip_address_udp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 80))
    ret = s.getsockname()[0]
    s.close()
    return ret


def is_open(ip, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        return True
    except:
        pass
    finally:
        s.close()
    return False



import logging
import logging.handlers
import socket
import os
import json

from cpbox.tool import net
from cpbox.tool import file
from cpbox.tool import timeutil
from cpbox.tool import system

class Holder():

    def __init__(self):
        self.basic_handlers = []
        pass

holder = Holder()

try:
    from threading import get_ident
except ImportError:
    from thread import get_ident

# This is the even log part we want: 2018-11-12T15:34:34+08:00 117.50.2.88 15051.50683 cp mid-call {"fail":0,"service":"id_service.next_id","rt_1":1.1670589447021,"rt":1.1730194091797,"env":"prod"}
# python strftime can not print timezone as +08:00
# logstash grok ISO8601_TIMEZONE will let this go: 2018-11-12T15:36:46+0800 172.16.1.150 28114.139798089959232 $appname test {"env": "dev", "time": "2018-11-12 15:36:46.359019"}
# ${time_iso_8601} ${client_ip} ${pid}.${rnd/thread_id} {$app_name} ${event_key} ${payload_json_encoded}

syslog_full_msg_pre = '%(time_iso_8601)s %(local_ip)s %(app_name)s_event_log[%(pid)d]: '
syslog_client_msg_pre = '%(app_name)s_event_log[%(pid)d]: '
basic_msg_pre = '%(time_iso_8601)s %(local_ip)s %(pid)d.%(tid)d '

readable_format = basic_msg_pre + '%(levelname)s %(name)s@%(filename)s:%(lineno)d %(message)s'

event_log_by_filelog_format = syslog_full_msg_pre + basic_msg_pre + '%(message)s'
clog_by_filelog_format = syslog_full_msg_pre + basic_msg_pre + '%(clog_message)s'

event_log_by_syslog_format = syslog_client_msg_pre + basic_msg_pre + '%(message)s'
clog_by_syslog_format = syslog_client_msg_pre + basic_msg_pre + '%(clog_message)s'

class BasicContextFilter(logging.Filter):

    def __init__(self):
        self.local_ip = net.get_local_ip_address()

    def filter(self, record):
        record.time_iso_8601 = timeutil.local_now_ios8061_str()
        record.local_ip = self.local_ip
        record.pid = os.getpid()
        record.tid = get_ident()
        return True

class ContextFilter(logging.Filter):

    def __init__(self, app):
        self.local_ip = net.get_local_ip_address()
        self.app = app

    def filter(self, record):
        record.time_iso_8601 = timeutil.local_now_ios8061_str()
        record.local_ip = self.local_ip
        record.app_name = self.app.app_name
        record.pid = os.getpid()
        record.tid = get_ident()
        return True

class ClogContextFilter(logging.Filter):

    def __init__(self, app):
        self.local_ip = net.get_local_ip_address()
        self.app = app

    def buildCLogMessage(self, record):
        clog_data = {}
        clog_data['env'] = self.app.env
        clog_data['app'] = self.app.app_name
        clog_data['pid'] = os.getpid()
        clog_data['tid'] = get_ident()

        clog_data['level'] = record.levelname
        clog_data['filename'] = record.filename
        clog_data['lineno'] = record.lineno
        clog_data['msg'] = record.getMessage()

        return 'cp clog %s' % json.dumps(clog_data)

    def filter(self, record):
        record.time_iso_8601 = timeutil.local_now_ios8061_str()
        record.app_name = self.app.app_name
        record.env = self.app.env
        record.local_ip = self.local_ip
        record.pid = os.getpid()
        record.tid = get_ident()
        record.clog_message = self.buildCLogMessage(record)
        return True

def getLogger(name=''):
    logger = logging.getLogger(name)
    return logger

def _get_rotating_filehandler(app, filename):
    file.ensure_dir_for_file(filename)
    handler = logging.handlers.RotatingFileHandler(filename, mode='a', maxBytes=1024 * 1024 * 128, backupCount=3)
    os.chmod(filename, 0o666)
    return handler

def _get_syslog_handler(syslog_ng_server):
    # SysLogHandler has no retry policy, in a long run script, server side will clost the connection which will cause the broken pipe error.
    # It's just a toy. We would like using UDP instead of TCP.
    return logging.handlers.SysLogHandler(address=(syslog_ng_server, 514), facility='local6')
    syslog_handler = None
    tcp_port = 601
    if net.is_open(syslog_ng_server, tcp_port):
        syslog_handler = logging.handlers.SysLogHandler(address=(syslog_ng_server, tcp_port), facility='local7', socktype=socket.SOCK_STREAM)
    else:
        syslog_handler = logging.handlers.SysLogHandler(address=(syslog_ng_server, 514), facility='local6')
    return syslog_handler

def _map_log_level(log_level_str):
    log_level = logging.INFO
    if log_level_str is not None:
        level = log_level_str.lower()
        log_level_conf = {
            'verbose': logging.NOTSET,
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'warn': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
        }
        if level in log_level_conf:
            log_level = log_level_conf[level]
    return log_level

def setup_readable_logger_handler(app, handler):
    formatter = logging.Formatter(readable_format)
    filter = ContextFilter(app)

    handler.setFormatter(formatter)
    handler.setLevel(_map_log_level(app.log_level))
    handler.addFilter(filter)
    return handler

def setup_clog_file_handler(app, handler, format):
    filter = ClogContextFilter(app)
    handler.setFormatter(logging.Formatter(format))
    handler.setLevel(_map_log_level(app.log_level))
    handler.addFilter(filter)
    return handler

def setup_event_log_handler(app, handler, format):
    filter = ContextFilter(app)
    handler.setFormatter(logging.Formatter(format))
    handler.addFilter(filter)

    event_logger = logging.getLogger('event-log')
    event_logger.propagate = False
    event_logger.setLevel(logging.INFO)
    event_logger.addHandler(handler)
    return event_logger


def make_logger_for_test():
    basic_handlers = []
    stream_handler = logging.StreamHandler()

    formatter = logging.Formatter(readable_format)
    filter = BasicContextFilter()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(_map_log_level('debug'))
    stream_handler.addFilter(filter)

    basic_handlers.append(stream_handler)

    root_logger = logging.getLogger()
    root_logger.handlers = []
    for h in basic_handlers:
        root_logger.setLevel(h.level)
        root_logger.addHandler(h)

def make_logger_for_app(app, **kwargs):
    log_config = kwargs.get('log_config', {})
    basic_handlers = []

    os_name = system.get_os_name()
    syslog_ng_server = os.environ['PUPPY_SYSLOG_NG_SERVER'] if 'PUPPY_SYSLOG_NG_SERVER' in os.environ else None
    if os_name == 'mac' and syslog_ng_server:
        event_log_syslog_handler = _get_syslog_handler(syslog_ng_server)
        setup_event_log_handler(app, event_log_syslog_handler, event_log_by_syslog_format)

        clog_syslog_handler = _get_syslog_handler(syslog_ng_server)
        setup_clog_file_handler(app, clog_syslog_handler, clog_by_syslog_format)
        basic_handlers.append(clog_syslog_handler)
    else:
        filename = app.app_logs_dir / 'syslog' / (app.app_name + '_event_log.log')
        event_log_file_handler = _get_rotating_filehandler(app, filename)
        setup_event_log_handler(app, event_log_file_handler, event_log_by_filelog_format)

        filename = app.app_logs_dir / 'syslog' / (app.app_name + '_clog_event_log.log')
        clog_file_handler = _get_rotating_filehandler(app, filename)
        setup_clog_file_handler(app, clog_file_handler, clog_by_filelog_format)
        basic_handlers.append(clog_file_handler)

    file_log = log_config.get('file_log', True)
    if file_log:
        filename = '%s/%s.log' % (app.app_logs_dir, app.app_name)
        app_filelog_handler = _get_rotating_filehandler(app, filename)
        setup_readable_logger_handler(app, app_filelog_handler)
        basic_handlers.append(app_filelog_handler)

    if True:
        import sys
        stream_handler = logging.StreamHandler(sys.stdout)
        setup_readable_logger_handler(app, stream_handler)
        basic_handlers.append(stream_handler)

    root_logger = logging.getLogger()
    root_logger.handlers = []
    for h in basic_handlers:
        root_logger.setLevel(h.level)
        root_logger.addHandler(h)
    holder.basic_handlers = basic_handlers

def shutdown():
    for handler in holder.basic_handlers:
        handler.close()



import json
import os

from backports import configparser
from ruamel.yaml import YAML

from cpbox.tool import array
from cpbox.tool import file

yaml = YAML(typ='rt')


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

def load_properties(fn, fallback=None, to_dict=True):
    if not os.path.isfile(fn):
        return fallback
    data = None
    config_string = '[DEFAULT]\n' + file.file_get_contents(fn)
    cp = configparser.ConfigParser()
    cp.optionxform = str
    cp.read_string(config_string)
    return dict(cp['DEFAULT']) if to_dict else cp['DEFAULT']

def dump_properties(data, fn):
    cp = configparser.ConfigParser()
    cp.optionxform = str
    cp['DEFAULT'] = data
    with open(fn, 'w') as f:
        cp.write(f, space_around_delimiters=False)
    file_content = file.file_get_contents(fn)
    file_content = file_content.replace('[DEFAULT]\n', '').strip()
    file.file_put_contents(fn, file_content)

def load_json(fn, fallback=None):
    if not os.path.isfile(fn):
        return fallback
    data = None
    with open(fn, 'r') as f:
        data = json.load(f)
    return data if data is not None else fallback

def load_yaml(fn, fallback=None):
    if not os.path.isfile(fn):
        return fallback
    data = None
    with open(fn, 'r') as f:
        data = yaml.load(f)
    return data if data is not None else fallback

def load_yaml_from_dir(dir, fallback=None):
    if not os.path.isdir(dir):
        return fallback
    file_list = file.listdir(dir)
    data = {}
    for path in file_list:
        if '.yml' not in path and '.yaml' not in path:
            continue
        data = array.merge(data, load_yaml(path, {}))
    return data

def load_yaml_config(config_file, config_section=None, extra_kvs=None):
    config = load_yaml(config_file, {})
    if config_section is not None:
        key_list = config_section.split('.')
        for sub_key in key_list:
            if sub_key not in config:
                raise Exception('Can not find config section for sub key:' +  sub_key)
            config = config[sub_key]

    if extra_kvs is not None:
        config.update(extra_kvs)
    return config

def dump_json(data, fn, **kwargs):
    with open(fn, 'w') as outfile:
        json.dump(data, outfile, **kwargs)

def pjson(data):
    return json.dumps(data, indent=2, sort_keys=True)

def pyaml(data):
    yaml.Dumper.ignore_aliases = lambda *args : True
    return yaml.dump(data)

# TODO: should be dump_yaml(data, fn)
def dump_yaml(fn, data):
    with open(fn, 'w') as outfile:
        yaml.dump(data, outfile)

def json_to_yaml(src, out):
    data = load_json(src)
    dump_yaml(data, out)



import unittest
from cpbox.tool import strings

class TestUtils(unittest.TestCase):

    def test_all(self):
        map = {
                'a': 'A',
                '1.1': '1.1',
                'scaling_group': 'ScalingGroup',
                }
        for src, dst in map.items():
            self.assertEqual(strings.camelize_key(src), dst)

if __name__ =='__main__':
    unittest.main()



from cpbox.tool import functocli
from cpbox.tool import file
from cpbox.tool import dockerutil
from cpbox.tool import template
from cpbox.app.env import EnvTool

import os
import sys

mysql_image = 'mysql:5.6'
busybox_image = 'busybox'


@functocli.keep_class
class MysqlCli(object):

    def __init__(self, app, mysql_config_dir, mysql_init_script_dir, mysql_data_dir, mysql_logs_dir, port=3306,
                 container_name='mysql', volumes=None):
        self.app = app
        self.mysql_config_dir = mysql_config_dir
        self.mysql_init_script_dir = mysql_init_script_dir
        self.mysql_data_dir = mysql_data_dir
        self.mysql_logs_dir = mysql_logs_dir
        self.port = port
        self.env_tool = EnvTool(self.app)
        self.mysql_image = self.env_tool.docker_url(mysql_image)
        self.busybox_image = self.env_tool.docker_url(busybox_image)
        self.mysql_container = container_name
        self.volumes = volumes

    def start_mysql(self):
        file.ensure_dir(self.mysql_data_dir)
        file.ensure_dir(self.mysql_logs_dir)
        ports = [
            str(self.port) + ':3306',
        ]

        volumes = {
            self.mysql_config_dir: '/etc/mysql/conf.d/',
            self.mysql_logs_dir: '/var/log/mysql',
            self.mysql_data_dir: '/var/lib/mysql',
        }
        if self.mysql_init_script_dir:
            volumes[self.mysql_init_script_dir] = '/docker-entrypoint-initdb.d'

        if self.volumes:
            volumes.update(self.volumes)

        envs = [
            'MYSQL_ROOT_PASSWORD=""',
            "MYSQL_ALLOW_EMPTY_PASSWORD=yes",
        ]

        args = dockerutil.base_docker_args(container_name=self.mysql_container, ports=ports, envs=envs, volumes=volumes)
        cmd_data = {'image': self.mysql_image, 'args': args}
        cmd = template.render_str('docker run -d --restart always {{ args }} {{ image }}', cmd_data)
        self.app.shell_run(cmd)
        self.wait_mysql()

    def wait_mysql(self):
        # cmd = 'while ! mysqladmin ping --silent; do sleep 1; done'
        cmd = 'while ! mysqladmin ping -h 127.0.0.1 --silent; do sleep 1; done'
        self.send_cmd_to_mysql_container(cmd)

    def send_cmd_to_mysql_container(self, cmd):
        cmd = 'docker exec %s %s bash -c \'%s\'' % (dockerutil.fg_mode(), self.mysql_container, cmd)
        self.app.shell_run(cmd)

    def exec_cmd_at_mysql_container(self, cmd):
        cmd = 'docker exec %s %s bash -c \'%s\'' % (dockerutil.fg_mode(), self.mysql_container, cmd)
        return self.app.run_cmd_ret(cmd)

    def to_mysql_env(self):
        cmd = 'bash'
        self.send_cmd_to_mysql_container(cmd)

    def to_mysql(self):
        cmd = 'mysql -A'
        self.send_cmd_to_mysql_container(cmd)

    def stop_mysql(self):
        self.app.stop_container(self.mysql_container)
        self.app.remove_container(self.mysql_container)

    def restart_mysql(self):
        self.stop_mysql()
        self.start_mysql()

    def rebuild_mysql(self):
        self.stop_mysql()
        cmd = 'rm -rf ' + self.mysql_data_dir
        self.sudo_for_storage(cmd)
        self.start_mysql()

    def sudo_for_storage(self, cmd):
        dir = os.environ.get('CPBOX_DATA_DIR', '/opt/data')
        cmd = 'docker run --rm %s -v %s:%s %s sh -c "%s"' % (dockerutil.fg_mode(), dir, dir, busybox_image, cmd)
        self.app.shell_run(cmd)



import unittest

from cpbox.tool import concurrent
from cpbox.tool import logger
from cpbox.tool import system


class SimpleThreadPoolTest:

    def __init__(self):
        self.count = 0
        self.total = 100

    def right(self):
        return self.count == self.total

    def run(self):
        def callback():
            self.count += 1
            system.shell_run('sleep 0.01')

        tp = concurrent.SimpleThreadPool()
        for i in range(0, self.total):
            tp.add_task(callback)
        tp.start()
        tp.wait()

class TestThreadPool(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_simple_thread_pool(self):
        test = SimpleThreadPoolTest()
        test.run()
        self.assertTrue(test.right())

if __name__ == '__main__':
    logger.make_logger_for_test()
    unittest.main()



from setuptools import setup, find_packages

setup(
    name = 'cpbox',
    version = '1.5.415',
    keywords = ('cpbox'),
    description = 'cp tool box',
    license = '',
    install_requires = [
        'six',
        'ruamel.yaml==0.18.5',
        'Jinja2',
        'netaddr',
        'requests',
        'tzlocal==2.1',
        'redis',
        'configparser==3.7.4',
        'oss2',
        ],

    scripts = [],

    author = 'https://www.liaohuqiu.com',
    author_email = 'liaohuqiu@gmail.com',
    url = '',

    packages = find_packages(),
    platforms = 'any',
)



CACHE_TYPE_EMPTY = 1
CACHE_TYPE_SIMPLE = 2

class EmptyCache():

    def __init__(self):
        self.type = CACHE_TYPE_EMPTY

    def has(self, key):
        return False

    def set(self, key, value, expire = 600):
        pass

    def get(self, key):
        return False

    def delete(self, key):
        pass

class SimpleCache():

    def __init__(self):
        self.type = CACHE_TYPE_SIMPLE
        self.cache_data = {}

    def has(self, key):
        return key in self.cache_data

    def set(self, key, value, expire = 600):
        self.cache_data[key] = value

    def get(self, key):
        if key in cache_data:
            return cache_data[key]
        return False

    def delete(self, key):
        if key in cache_data:
            del cache_data[key]

empty_cache = EmptyCache()
simple_cache = SimpleCache()

cache_proxy = empty_cache
def set_cach_proxy(proxy):
    cache_proxy = proxy

def has(key):
    return cache_proxy.has(key)

def set(key, value, expire = 600):
    return cache_proxy.set(key, value, expire)

def get(key):
    return cache_proxy.get(key)

def delete(key):
    return cache_proxy.delete(key)
